@@ -10,14 +10,14 @@ pydot Homepage:	http://www.dkbza.org/pydot.html
 Graphviz:	http://www.research.att.com/sw/tools/graphviz/
 DOT Language:	http://www.research.att.com/~erg/graphviz/info/lang.html
 
-Programmed and tested with Graphviz 1.12 and Python 2.3.3 on GNU/Linux
+Programmed and tested with Graphviz 1.16 and Python 2.3.4 on GNU/Linux
 by Ero Carrera (c) 2004	[ero@dkbza.org]
 
 Distributed under MIT license [http://opensource.org/licenses/mit-license.html].
 """
 
 __author__ = 'Ero Carrera'
-__version__ = '0.9.9'
+__version__ = '0.9.10'
 __license__ = 'MIT'
 
 import os
@@ -214,7 +214,7 @@ class Node(object, Common):
 		'distortion', 'fixedsize', 'group', 'height', 'orientation', 'pin',	\
 		'rects', 'regular', 'shape', 'shapefile', 'sides', 'skew', 'pos',	\
 		'layer', 'tooltip', 'style', 'target', 'color', 'peripheries',
-		'fillcolor']
+		'fillcolor', 'margin', 'nojustify']
 
 	def __init__(self, name, **attrs):
 	
@@ -230,10 +230,37 @@ class Node(object, Common):
 			# Generate all the Setter methods.
 			self.__setattr__('set_'+attr, lambda x, a=attr : self.__setattr__(a, x))
 			# Generate all the Getter methods.
-			self.__setattr__('get_'+attr, lambda a=attr : self.__getattribute__(a))
+			self.__setattr__('get_'+attr, lambda a=attr : self.__get_attribute__(a))
 		for attr, val in attrs.items():
 			self.__setattr__(attr, val)
 
+	def __getstate__(self):
+
+		dict = copy.copy(self.__dict__)
+		for attr in self.attributes:
+			del dict['set_'+attr]
+			del dict['get_'+attr]
+   
+		return dict
+	
+	def __setstate__(self, state):
+		for k, v in state.items():
+			self.__setattr__(k, v)
+	
+	def __get_attribute__(self, attr):
+		"""Look for default attributes for this node"""
+		attr_val = self.__getattribute__(attr)
+		if attr_val is None:
+			defaults = self.parent_graph.get_node('node')
+			if defaults:
+				attr_val = defaults.__getattribute__(attr)
+				if attr_val:
+					return attr_val
+		else:
+			return attr_val
+		return None
+	
+	
 	def set_name(self, node_name):
 		"""Set the node's name."""
 		
@@ -266,7 +293,7 @@ class Node(object, Common):
 			self.name = str(self.name)
 		
 		# RMF: special case defaults for node, edge and graph properties.
-		if self.name in ['node', 'edge', 'graph']:
+		if self.name in ['node', 'edge', 'graph'] or self.name.startswith('"'):
 			node = self.name
 		else:
 			node = '"'+self.name+'"'
@@ -283,11 +310,13 @@ class Node(object, Common):
 				else:
 					node_attr += ', '
 				node_attr += attr+'='
-				val = self.__dict__[attr]
+				val = str(self.__dict__[attr])
 				
 				if val.startswith('<') and val.endswith('>'):
 					node_attr += val
-				elif (isinstance(val, str) or isinstance(val, unicode)) and not self.is_ID(val):
+				elif ((isinstance(val, str) or isinstance(val, unicode)) and \
+					not self.is_ID(val)) or val == '' :
+					
 					node_attr += '"'+val+'"'
 				else:
 					node_attr += str(val)
@@ -328,7 +357,7 @@ class Edge(object, Common):
 		'labelfloat', 'labelfontcolor', 'labelfontname', 'labelfontsize', 'len',\
 		'lhead', 'ltail', 'minlen', 'samehead', 'sametail', 'weight', 'tailURL',\
 		'tailclip', 'tailhref', 'taillabel', 'tailport', 'tailtarget',		\
-		'tailtooltip']
+		'tailtooltip', 'nojustify']
 
 	def __init__(self, src, dst, **attrs):
 		self.src = src
@@ -339,10 +368,37 @@ class Edge(object, Common):
 			# Generate all the Setter methods.
 			self.__setattr__('set_'+attr, lambda x, a=attr : self.__setattr__(a, x))
 			# Generate all the Getter methods.
-			self.__setattr__('get_'+attr, lambda a=attr : self.__getattribute__(a))
+			self.__setattr__('get_'+attr, lambda a=attr : self.__get_attribute__(a))
 		for attr, val in attrs.items():
 			self.__setattr__(attr, val)
 
+	def __getstate__(self):
+
+		dict = copy.copy(self.__dict__)
+		for attr in self.attributes:
+			del dict['set_'+attr]
+			del dict['get_'+attr]
+
+		return dict
+	
+	def __setstate__(self, state):
+		for k, v in state.items():
+			self.__setattr__(k, v)
+	
+	def __get_attribute__(self, attr):
+		"""Look for default attributes for this edge"""
+		attr_val = self.__getattribute__(attr)
+		if attr_val is None:
+			defaults = self.parent_graph.get_node('edge')
+			if defaults:
+				attr_val = defaults.__getattribute__(attr)
+				if attr_val:
+					return attr_val
+		else:
+			return attr_val
+		return None
+	
+	
 	def get_source(self):
 		"""Get the edges source node name."""
 	
@@ -449,7 +505,7 @@ class Edge(object, Common):
 				else:
 					edge_attr+=', '
 				edge_attr+=attr+'='
-				val = self.__dict__[attr]
+				val = str(self.__dict__[attr])
 				if (isinstance(val, str) or isinstance(val, unicode)) and not self.is_ID(val):
 					edge_attr+='"'+val+'"'
 				else:
@@ -504,7 +560,7 @@ class Graph(object, Common):
 		'viewport', 'voro_margin', 'quantum', 'bgcolor', 'labeljust',		\
 		'labelloc', 'root', 'showboxes', 'URL', 'fontcolor', 'fontsize',	\
 		'label' ,'fontname', 'comment', 'lp', 'target', 'color', 'style',	\
-		'concentrators']
+		'concentrators', 'rank', 'dpi', 'mode', 'nojustify', 'nslimit']
 
 	def __init__(self, graph_name='G', graph_type='digraph', strict=False,	\
 		suppress_disconnected=False, simplify=False, **attrs):
@@ -521,6 +577,7 @@ class Graph(object, Common):
 		self.edge_src_list = []
 		self.edge_dst_list = []
 		self.subgraph_list = []
+		self.sorted_graph_elements = []
 		self.parent_graph = self
 		for attr in self.attributes:
 			# Set all the attributes to None.
@@ -528,10 +585,36 @@ class Graph(object, Common):
 			# Generate all the Setter methods.
 			self.__setattr__('set_'+attr, lambda x, a=attr : self.__setattr__(a, x))
 			# Generate all the Getter methods.
-			self.__setattr__('get_'+attr, lambda a=attr : self.__getattribute__(a))
+			self.__setattr__('get_'+attr, lambda a=attr : self.__get_attribute__(a))
 		for attr, val in attrs.items():
 			self.__setattr__(attr, val)
 	
+	def __getstate__(self):
+
+		dict = copy.copy(self.__dict__)
+		for attr in self.attributes:
+			del dict['set_'+attr]
+			del dict['get_'+attr]
+			
+		return dict
+		
+	def __setstate__(self, state):
+		for k, v in state.items():
+			self.__setattr__(k, v)
+			
+	def __get_attribute__(self, attr):
+		"""Look for default attributes for this graph"""
+		attr_val = self.__getattribute__(attr)
+		if attr_val is None:
+			defaults = self.get_node('graph')
+			if defaults:
+				attr_val = defaults.__getattribute__(attr)
+				if attr_val:
+					return attr_val
+		else:
+			return attr_val
+		return None
+
 	def set_simplify(self, simplify):
 		"""Set whether to simplify or not.
 		
@@ -641,9 +724,16 @@ class Graph(object, Common):
 		if not isinstance(graph_node, Node):
 			raise Error, 'add_node received a non node class object'
 			
-		if self.get_node(graph_node.get_name()) is None:
-			self.node_list.append(graph_node)
-			graph_node.parent_graph = self.parent_graph
+		node = self.get_node(graph_node.get_name())
+		if node is None or graph_node.get_name() in ('graph', 'node', 'edge'):
+				self.node_list.append(graph_node)
+				graph_node.parent_graph = self.parent_graph
+		elif (node.__dict__.has_key('added_from_edge') and node.added_from_edge):
+			for k, v in graph_node.__dict__.items():
+				if v is not None and node.__dict__.has_key(k) and node.__dict__[k] is None:
+					node.__dict__[k] = v
+					
+		self.sorted_graph_elements.append(graph_node)
 
 	def get_node(self, name):
 		"""Retrieved a node from the graph.
@@ -674,7 +764,7 @@ class Graph(object, Common):
 		composing the graph.
 		"""
 		
-		return self.node_list
+		return [n for n in self.node_list if n.get_name() not in ('graph', 'edge', 'node')]
 			
 	def add_edge(self, graph_edge):
 		"""Adds an edge object to the graph.
@@ -690,11 +780,11 @@ class Graph(object, Common):
 		
 		src = self.get_node(graph_edge.get_source())
 		if src is None:
-			self.add_node(Node(graph_edge.get_source()))
+			self.add_node(Node(graph_edge.get_source(), added_from_edge=True))
 		
 		dst = self.get_node(graph_edge.get_destination())
 		if dst is None:
-			self.add_node(Node(graph_edge.get_destination()))
+			self.add_node(Node(graph_edge.get_destination(), added_from_edge=True))
 		
 		graph_edge.parent_graph = self.parent_graph
 		
@@ -703,6 +793,8 @@ class Graph(object, Common):
 			
 		if graph_edge.dst not in self.edge_dst_list:
 			self.edge_dst_list.append(graph_edge.dst)
+			
+		self.sorted_graph_elements.append(graph_edge)
 
 	def get_edge(self, src, dst):
 		"""Retrieved an edge from the graph.
@@ -747,6 +839,9 @@ class Graph(object, Common):
 		self.subgraph_list.append(sgraph)
 		
 		sgraph.set_graph_parent(self.parent_graph)
+		
+		self.sorted_graph_elements.append(sgraph)
+		
 		return None
 	
 	def get_subgraph(self, name):
@@ -798,12 +893,12 @@ class Graph(object, Common):
 			elm.parent_graph = parent
 			elm.set_graph_parent(parent)
 		
-	def to_string(self, indent=''):
+	def to_string(self):
 		"""Returns a string representation of the graph in dot language.
 		
 		It will return the graph and all its subelements in string from.
 		"""
-		graph = indent+''
+		graph = ''
 		if self.__dict__.has_key('strict'):
 			if self==self.parent_graph and self.strict:
 				graph+='strict '
@@ -813,41 +908,31 @@ class Graph(object, Common):
 		for attr in self.attributes:
 			if self.__dict__.has_key(attr)	\
 			and self.__getattribute__(attr) is not None:
-				graph += indent+'\t'+attr+'='
-				val = self.__dict__[attr]
+				graph += attr+'='
+				val = str(self.__dict__[attr])
 				if isinstance(val, str) and not self.is_ID(val):
 					graph += '"'+val+'"'
 				else:
 					graph += str(val)
 				graph+=';\n'
 
-		# RMF: for graphviz to handle edge/node defaults correctly,
-		# it seems they must appear before the rest of the nodes/edges
-		# so if they exist, dump them out first.
-		for elm in self.node_list:
-			if elm.name in ['node', 'edge', 'graph']:
-				graph += indent+'\t'+elm.to_string()+'\n'
-		
-		for elm in self.subgraph_list:
-			graph += elm.to_string(indent+'\t')+'\n'
-				
-		for elm in self.node_list:
-			# RMF: don't repeat the node/edge defaults
-			if elm.name in ['node', 'edge', 'graph']:
-				continue
-			if self.suppress_disconnected:
-				if elm.name not in self.edge_src_list+self.edge_dst_list:
-					continue
-			graph+=indent+'\t'+elm.to_string()+'\n'
 
 		edges_done = []
-		for elm in self.edge_list:
-			if self.simplify and elm in edges_done:
-				continue
-			graph+=indent+'\t'+elm.to_string()+'\n'
-			edges_done.append(elm)
+		for elm in self.sorted_graph_elements:
+			if isinstance(elm, Node):
+				if self.suppress_disconnected:
+					if elm.name not in self.edge_src_list+self.edge_dst_list:
+						continue
+				graph += elm.to_string()+'\n'
+			elif isinstance(elm, Edge):
+				if self.simplify and elm in edges_done:
+					continue
+				graph += elm.to_string()+'\n'
+				edges_done.append(elm)
+			else:
+				graph += elm.to_string()+'\n'
 
-		graph+=indent+'}\n'
+		graph += '}\n'
 		
 		return graph
 
@@ -879,7 +964,7 @@ class Subgraph(Graph):
 	 	subgraph_instance.label, subgraph_instance.fontname
 	"""
 	
-	attributes = Graph.attributes+['rank']
+	attributes = Graph.attributes
 	
 	# RMF: subgraph should have all the attributes of graph so it can be passed
 	# as a graph to all methods
@@ -895,16 +980,43 @@ class Subgraph(Graph):
 		self.edge_src_list = []
 		self.edge_dst_list = []
 		self.subgraph_list = []
+		self.sorted_graph_elements = []
 		for attr in self.attributes:
 			# Set all the attributes to None.
 			self.__setattr__(attr, None)
 			# Generate all the Setter methods.
 			self.__setattr__('set_'+attr, lambda x, a=attr : self.__setattr__(a, x))
 			# Generate all the Getter methods.
-			self.__setattr__('get_'+attr, lambda a=attr : self.__getattribute__(a))
+			self.__setattr__('get_'+attr, lambda a=attr : self.__get_attribute__(a))
 		for attr, val in attrs.items():
 			self.__setattr__(attr, val)
-			
+	
+	def __getstate__(self):
+
+		dict = copy.copy(self.__dict__)
+		for attr in self.attributes:
+			del dict['set_'+attr]
+			del dict['get_'+attr]
+   
+		return dict
+	
+	def __setstate__(self, state):
+		for k, v in state.items():
+			self.__setattr__(k, v)
+	
+	def __get_attribute__(self, attr):
+		"""Look for default attributes for this subgraph"""
+		attr_val = self.__getattribute__(attr)
+		if attr_val is None:
+			defaults = self.get_node('graph')
+			if defaults:
+				attr_val = defaults.__getattribute__(attr)
+				if attr_val:
+					return attr_val
+		else:
+			return attr_val
+		return None
+
 
 class Cluster(Graph):
 
@@ -935,7 +1047,7 @@ class Cluster(Graph):
 	
 	attributes = ['pencolor', 'bgcolor', 'labeljust', 'labelloc', 'URL', 'fontcolor',	\
 		'fontsize', 'label', 'fontname', 'lp', 'style', 'target', 'color',	\
-		'peripheries', 'fillcolor']
+		'peripheries', 'fillcolor', 'nojustify']
 
 	def __init__(self, graph_name='subG', suppress_disconnected=False,	\
 		simplify=False, **attrs):
@@ -951,15 +1063,42 @@ class Cluster(Graph):
 		self.edge_src_list = []
 		self.edge_dst_list = []
 		self.subgraph_list = []
+		self.sorted_graph_elements = []
 		for attr in self.attributes:
 			# Set all the attributes to None.
 			self.__setattr__(attr, None)
 			# Generate all the Setter methods.
 			self.__setattr__('set_'+attr, lambda x, a=attr : self.__setattr__(a, x))
 			# Generate all the Getter methods.
-			self.__setattr__('get_'+attr, lambda a=attr : self.__getattribute__(a))
+			self.__setattr__('get_'+attr, lambda a=attr : self.__get_attribute__(a))
 		for attr, val in attrs.items():
 			self.__setattr__(attr, val)
+	
+	def __getstate__(self):
+
+		dict = copy.copy(self.__dict__)
+		for attr in self.attributes:
+			del dict['set_'+attr]
+			del dict['get_'+attr]
+   
+		return dict
+	
+	def __setstate__(self, state):
+		for k, v in state.items():
+			self.__setattr__(k, v)
+	
+	def __get_attribute__(self, attr):
+		"""Look for default attributes for this cluster"""
+		attr_val = self.__getattribute__(attr)
+		if attr_val is None:
+			defaults = self.get_node('graph')
+			if defaults:
+				attr_val = defaults.__getattribute__(attr)
+				if attr_val:
+					return attr_val
+		else:
+			return attr_val
+		return None
 
 
 class Dot(Graph):
@@ -973,7 +1112,7 @@ class Dot(Graph):
 	progs = None
 	
 	formats = ['ps', 'ps2', 'hpgl', 'pcl', 'mif', 'pic', 'gd', 'gd2', 'gif', 'jpg',	\
-		'jpeg', 'png', 'wbmp', 'ismap', 'imap', 'cmap', 'vrml', 'vtx', 'mp',	\
+		'jpeg', 'png', 'wbmp', 'ismap', 'imap', 'cmap', 'cmapx', 'vrml', 'vtx', 'mp',	\
 		'fig', 'svg', 'svgz', 'dia', 'dot', 'canon', 'plain', 'plain-ext', 'xdot']
 
 	def __init__(self, prog='dot', **args):
@@ -992,7 +1131,27 @@ class Dot(Graph):
 			self.__setattr__('write_'+frmt, lambda path, f=frmt, prog=self.prog : self.write(path, format=f, prog=prog))
 			f = self.__dict__['write_'+frmt]
 			f.__doc__ = '''Refer to docstring from 'write' for more information.'''
+
+			
+	def __getstate__(self):
+
+		dict = copy.copy(self.__dict__)
+		for attr in self.attributes:
+			del dict['set_'+attr]
+			del dict['get_'+attr]
+   			
+		for k in [x for x in dict.keys() if x.startswith('write_')] + \
+			[x for x in dict.keys() if x.startswith('create_')]:
+			del dict[k]
+   
+		return dict
 	
+	def __setstate__(self, state):
+		self.__init__()
+		for k, v in state.items():
+			self.__setattr__(k, v)
+
+				
 	def set_prog(self, prog):
 		"""Sets the default program.
 		
